@@ -98,8 +98,8 @@ static void dump_boot(DOS_FS *fs,struct boot_sector *b,unsigned lss)
     }
     printf("Data area starts at byte %llu (sector %llu)\n",fs->data_start,
 	   fs->data_start/lss);
-    printf("%10lu data clusters (%lu bytes)\n",fs->clusters,fs->clusters*
-      fs->cluster_size);
+    printf("%10lu data clusters (%llu bytes)\n",fs->clusters,
+	   (unsigned long long)fs->clusters*fs->cluster_size);
     printf("%u sectors/track, %u heads\n",CF_LE_W(b->secs_track),
 	   CF_LE_W(b->heads));
     printf("%10u hidden sectors\n",
@@ -269,7 +269,8 @@ void read_boot(DOS_FS *fs)
 {
     struct boot_sector b;
     unsigned total_sectors;
-    unsigned short logical_sector_size, fat_length, sectors;
+    unsigned short logical_sector_size, sectors;
+    unsigned fat_length;
     loff_t data_size;
 
     fs_read(0,sizeof(b),&b);
@@ -277,10 +278,15 @@ void read_boot(DOS_FS *fs)
     if (!logical_sector_size) die("Logical sector size is zero.");
     fs->cluster_size = b.cluster_size*logical_sector_size;
     if (!fs->cluster_size) die("Cluster size is zero.");
-    if (b.fats != 2)
-	die("Currently, only 2 FATs are supported, not %d.\n",b.fats);
+    if (b.fats != 2 && b.fats != 1)
+	die("Currently, only 1 or 2 FATs are supported, not %d.\n",b.fats);
+    fs->nfats = b.fats;
     sectors = GET_UNALIGNED_W(b.sectors);
     total_sectors = sectors ? sectors : CF_LE_L(b.total_sect);
+    if (verbose) printf("Checking we can access the last sector of the filesystem\n");
+    /* Can't access last odd sector anyway, so round down */
+    fs_test((loff_t)((total_sectors & ~1)-1)*(loff_t)logical_sector_size,
+	    logical_sector_size);
     fat_length = CF_LE_W(b.fat_length) ?
 		 CF_LE_W(b.fat_length) : CF_LE_L(b.fat32_length);
     fs->fat_start = (loff_t)CF_LE_W(b.reserved)*logical_sector_size;
@@ -295,7 +301,6 @@ void read_boot(DOS_FS *fs)
     fs->fsinfo_start = 0; /* no FSINFO structure */
     fs->free_clusters = -1; /* unknown */
     if (!b.fat_length && b.fat32_length) {
-	fprintf( stderr, "Warning: FAT32 support is still ALPHA.\n" );
 	fs->fat_bits = 32;
 	fs->root_cluster = CF_LE_L(b.root_cluster);
 	if (!fs->root_cluster && fs->root_entries)
@@ -342,9 +347,9 @@ void read_boot(DOS_FS *fs)
     /* On FAT32, the high 4 bits of a FAT entry are reserved */
     fs->eff_fat_bits = (fs->fat_bits == 32) ? 28 : fs->fat_bits;
     fs->fat_size = fat_length*logical_sector_size;
-    if (fs->clusters > (fs->fat_size*8/fs->fat_bits)-2)
+    if (fs->clusters > ((unsigned long long)fs->fat_size*8/fs->fat_bits)-2)
 	die("File system has %d clusters but only space for %d FAT entries.",
-	  fs->clusters,(fs->fat_size*8/fs->fat_bits)-2);
+	  fs->clusters,((unsigned long long)fs->fat_size*8/fs->fat_bits)-2);
     if (!fs->root_entries && !fs->root_cluster)
 	die("Root directory has zero size.");
     if (fs->root_entries & (MSDOS_DPS-1))
