@@ -59,11 +59,15 @@ void read_fat(DOS_FS *fs)
 
     eff_size = ((fs->clusters+2)*fs->fat_bits+7)/8;
     first = alloc(eff_size);
-    second = alloc(eff_size);
     fs_read(fs->fat_start,eff_size,first);
-    fs_read(fs->fat_start+fs->fat_size,eff_size,second);
     use = first;
-    if (memcmp(first,second,eff_size) != 0) {
+    if (fs->nfats > 1) {
+	second = alloc(eff_size);
+	fs_read(fs->fat_start+fs->fat_size,eff_size,second);
+    }
+    else
+	second = NULL;
+    if (second && memcmp(first,second,eff_size) != 0) {
 	FAT_ENTRY first_media, second_media;
 	get_fat(&first_media,first,0,fs);
 	get_fat(&second_media,second,0,fs);
@@ -106,7 +110,8 @@ void read_fat(DOS_FS *fs)
 	    set_fat(fs,i,-1);
 	}
     free(first);
-    free(second);
+    if (second)
+	free(second);
 }
 
 
