@@ -268,6 +268,7 @@ char dummy_boot_jump[3] = { 0xeb, 0x3c, 0x90 };
 
 char dummy_boot_jump_m68k[2] = { 0x60, 0x1c };
 
+#define MSG_OFFSET_OFFSET 3
 char dummy_boot_code[BOOTCODE_SIZE] =
   "\x0e"			/* push cs */
   "\x1f"			/* pop ds */
@@ -561,23 +562,24 @@ count_blocks (char *filename)
       perror (filename);
       exit (1);
     }
-  low = 0;
 
-  for (high = 1; valid_offset (fd, high); high *= 2)
-    low = high;
-  while (low < high - 1)
-    {
-      const loff_t mid = (low + high) / 2;
+  /* first try SEEK_END, which should work on most devices nowadays */
+  if ((low = llseek(fd, 0, SEEK_END)) <= 0) {
+      low = 0;
+      for (high = 1; valid_offset (fd, high); high *= 2)
+	  low = high;
+      while (low < high - 1) {
+	  const loff_t mid = (low + high) / 2;
+	  if (valid_offset (fd, mid))
+	      low = mid;
+	  else
+	      high = mid;
+      }
+      ++low;
+  }
 
-      if (valid_offset (fd, mid))
-	low = mid;
-      else
-	high = mid;
-    }
-  valid_offset (fd, 0);
   close (fd);
-
-  return (low + 1) / BLOCK_SIZE;
+  return low / BLOCK_SIZE;
 }
 
 
@@ -770,13 +772,24 @@ establish_params (int device_num,int size)
       bs.media = (char) 0xf8; /* Set up the media descriptor for a hard drive */
       bs.dir_entries[0] = (char) 0;	/* Default to 512 entries */
       bs.dir_entries[1] = (char) 2;
-      /* For FAT32, use 4k clusters on sufficiently large file systems,
-       * otherwise 1 sector per cluster. This is also what M$'s format
-       * command does for FAT32. */
-      bs.cluster_size = (char)
-	    (size_fat == 32 ?
-	     ((long long)blocks*SECTORS_PER_BLOCK >= 512*1024 ? 8 : 1) :
-	     4); /* FAT12 and FAT16: start at 4 sectors per cluster */
+      if (size_fat == 32) {
+	  /* For FAT32, try to do the same as M$'s format command:
+	   * fs size < 256M: 0.5k clusters
+	   * fs size <   8G: 4k clusters
+	   * fs size <  16G: 8k clusters
+	   * fs size >= 16G: 16k clusters
+	   */
+	  unsigned long sz_mb =
+	      (blocks+(1<<(20-BLOCK_SIZE_BITS))-1) >> (20-BLOCK_SIZE_BITS);
+	  bs.cluster_size = sz_mb >= 16*1024 ? 32 :
+			    sz_mb >=  8*1024 ? 16 :
+			    sz_mb >=     256 ?  8 :
+					        1;
+      }
+      else {
+	  /* FAT12 and FAT16: start at 4 sectors per cluster */
+	  bs.cluster_size = (char) 4;
+      }
     }
 }
 
@@ -840,10 +853,14 @@ setup_tables (void)
 		       (char *)&bs) - 2;
 
     if (size_fat == 32) {
+	int offset = (char *)&bs.fat32.boot_code -
+		     (char *)&bs + MESSAGE_OFFSET + 0x7c00;
 	if (dummy_boot_code[BOOTCODE_FAT32_SIZE-1])
 	  printf ("Warning: message too long; truncated\n");
 	dummy_boot_code[BOOTCODE_FAT32_SIZE-1] = 0;
 	memcpy(bs.fat32.boot_code, dummy_boot_code, BOOTCODE_FAT32_SIZE);
+	bs.fat32.boot_code[MSG_OFFSET_OFFSET] = offset & 0xff;
+	bs.fat32.boot_code[MSG_OFFSET_OFFSET+1] = offset >> 8;
     }
     else {
 	memcpy(bs.oldfat.boot_code, dummy_boot_code, BOOTCODE_SIZE);
